"""
Test autosigning minions based on grain values.
"""

import os
import shutil
import stat
import subprocess

import pytest

import salt.utils.files
import salt.utils.stringutils
from tests.support.runtests import RUNTIME_VARS

pytestmark = [
    pytest.mark.slow_test,
]


@pytest.fixture
def autosign_file_permissions():
    return stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR


@pytest.fixture
def autosign_file_path(tmp_path):
    return str(tmp_path / "autosign_file")


@pytest.fixture
def autosign_grains_dir(salt_master):
    return salt_master.config["autosign_grains_dir"]


@pytest.fixture(autouse=True)
def setup_autosign_files(
    autosign_file_permissions,
    autosign_file_path,
    autosign_grains_dir,
    salt_key_cli,
    salt_call_cli,
    salt_minion,
):
    shutil.copyfile(
        os.path.join(RUNTIME_VARS.FILES, "autosign_grains", "autosign_file"),
        autosign_file_path,
    )
    os.chmod(autosign_file_path, autosign_file_permissions)

    salt_key_cli.run("-d", salt_minion.id, "-y")
    salt_call_cli.run(
        "test.ping", "-l", "quiet"
    )  # get minion to try to authenticate itself again

    skip_msg = ""
    if salt_minion.id in salt_key_cli.run("-l", "acc").data:
        skip_msg = "Could not deauthorize minion"
    if salt_minion.id not in salt_key_cli.run("-l", "un").data:
        skip_msg = "minion did not try to reauthenticate itself"

    if not os.path.isdir(autosign_grains_dir):
        os.makedirs(autosign_grains_dir)

    if not skip_msg:
        yield

    shutil.copyfile(
        os.path.join(RUNTIME_VARS.FILES, "autosign_file"), autosign_file_path
    )
    os.chmod(autosign_file_path, autosign_file_permissions)

    salt_call_cli.run(
        "test.ping", "-l", "quiet"
    )  # get minion to authenticate itself again

    try:
        if os.path.isdir(autosign_grains_dir):
            shutil.rmtree(autosign_grains_dir)
    except AttributeError:
        pass

    if skip_msg:
        pytest.skip(skip_msg)


@pytest.mark.slow_test
def test_autosign_grains_accept(
    autosign_grains_dir,
    autosign_file_permissions,
    salt_minion,
    salt_call_cli,
    salt_key_cli,
):
    grain_file_path = os.path.join(autosign_grains_dir, "test_grain")
    with salt.utils.files.fopen(grain_file_path, "w") as f:
        f.write(salt.utils.stringutils.to_str("#invalid_value\ncheese"))
    os.chmod(grain_file_path, autosign_file_permissions)

    salt_call_cli.run(
        "test.ping", "-l", "quiet"
    )  # get minion to try to authenticate itself again
    assert salt_minion.id in salt_key_cli.run("-l", "acc")


@pytest.mark.slow_test
def test_autosign_grains_fail(
    autosign_grains_dir,
    autosign_file_permissions,
    salt_minion,
    salt_call_cli,
    salt_key_cli,
):
    grain_file_path = os.path.join(autosign_grains_dir, "test_grain")
    with salt.utils.files.fopen(grain_file_path, "w") as f:
        f.write(salt.utils.stringutils.to_str("#cheese\ninvalid_value"))
    os.chmod(grain_file_path, autosign_file_permissions)

    salt_call_cli.run(
        "test.ping", "-l", "quiet"
    )  # get minion to try to authenticate itself again
    assert salt_minion.id not in salt_key_cli.run("-l", "acc")
    assert salt_minion.id in salt_key_cli.run("-l", "un")


@pytest.mark.skip_unless_on_linux
@pytest.mark.slow_test
def test_ufw_allow(salt_master, grains):
    if grains["os_family"] != "Debian":
        pytest.skip("Only runs on Debian family.")

    expected_output = """Skipping adding existing rule
Skipping adding existing rule (v6)
    """

    proc = subprocess.Popen(
        "ufw allow salt\n", shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE
    )
    out, err = proc.communicate()
    out_strg = out.decode()
    err_strg = err.decode()
    assert out_strg == expected_output
    assert err_strg != "ERROR: Could not find a profile matching 'salt'"



"""
Test for ssh_pre_flight roster option
"""

try:
    import grp
    import pwd
except ImportError:
    # windows stacktraces on import of these modules
    pass
import os
import pathlib
import shutil
import subprocess

import pytest
import yaml
from saltfactories.utils import random_string

import salt.utils.files

pytestmark = [
    pytest.mark.skip_on_windows(reason="Salt-ssh not available on Windows"),
]


def _custom_roster(roster_file, roster_data):
    with salt.utils.files.fopen(roster_file, "r") as fp:
        data = salt.utils.yaml.safe_load(fp)
    for key, item in roster_data.items():
        data["localhost"][key] = item
    with salt.utils.files.fopen(roster_file, "w") as fp:
        yaml.safe_dump(data, fp)


@pytest.fixture
def _create_roster(salt_ssh_roster_file, tmp_path):
    thin_dir = tmp_path / "thin-dir"
    ret = {
        "roster": salt_ssh_roster_file,
        "data": {
            "ssh_pre_flight": str(tmp_path / "ssh_pre_flight.sh"),
        },
        "test_script": str(tmp_path / "test-pre-flight-script-worked.txt"),
        "thin_dir": str(thin_dir),
    }

    with salt.utils.files.fopen(salt_ssh_roster_file, "r") as fp:
        data = salt.utils.yaml.safe_load(fp)

    pre_flight_script = ret["data"]["ssh_pre_flight"]
    data["localhost"]["ssh_pre_flight"] = pre_flight_script
    data["localhost"]["thin_dir"] = ret["thin_dir"]
    with salt.utils.files.fopen(salt_ssh_roster_file, "w") as fp:
        yaml.safe_dump(data, fp)

    with salt.utils.files.fopen(pre_flight_script, "w") as fp:
        fp.write("touch {}".format(ret["test_script"]))

    try:
        yield ret
    finally:
        if thin_dir.exists():
            shutil.rmtree(thin_dir)


@pytest.mark.slow_test
def test_ssh_pre_flight(salt_ssh_cli, caplog, _create_roster):
    """
    test ssh when ssh_pre_flight is set ensure the script runs successfully
    """
    ret = salt_ssh_cli.run("test.ping")
    assert ret.returncode == 0

    assert pathlib.Path(_create_roster["test_script"]).exists()


@pytest.mark.slow_test
def test_ssh_run_pre_flight(salt_ssh_cli, _create_roster):
    """
    test ssh when --pre-flight is passed to salt-ssh to ensure the script runs successfully
    """
    # make sure we previously ran a command so the thin dir exists
    ret = salt_ssh_cli.run("test.ping")
    assert pathlib.Path(_create_roster["test_script"]).exists()

    # Now remeove the script to ensure pre_flight doesn't run
    # without --pre-flight
    pathlib.Path(_create_roster["test_script"]).unlink()

    assert salt_ssh_cli.run("test.ping").returncode == 0
    assert not pathlib.Path(_create_roster["test_script"]).exists()

    # Now ensure
    ret = salt_ssh_cli.run("test.ping", "--pre-flight")
    assert ret.returncode == 0
    assert pathlib.Path(_create_roster["test_script"]).exists()


@pytest.mark.slow_test
def test_ssh_run_pre_flight_args(salt_ssh_cli, _create_roster):
    """
    test ssh when --pre-flight is passed to salt-ssh
    to ensure the script runs successfully passing some args
    """
    _custom_roster(salt_ssh_cli.roster_file, {"ssh_pre_flight_args": "foobar test"})
    # Create pre_flight script that accepts args
    test_script = _create_roster["test_script"]
    test_script_1 = pathlib.Path(test_script + "-foobar")
    test_script_2 = pathlib.Path(test_script + "-test")
    with salt.utils.files.fopen(_create_roster["data"]["ssh_pre_flight"], "w") as fp:
        fp.write(
            f"""
        touch {str(test_script)}-$1
        touch {str(test_script)}-$2
        """
        )
    ret = salt_ssh_cli.run("test.ping")
    assert ret.returncode == 0
    assert test_script_1.exists()
    assert test_script_2.exists()
    test_script_1.unlink()
    test_script_2.unlink()

    ret = salt_ssh_cli.run("test.ping")
    assert ret.returncode == 0
    assert not test_script_1.exists()
    assert not test_script_2.exists()

    ret = salt_ssh_cli.run("test.ping", "--pre-flight")
    assert ret.returncode == 0
    assert test_script_1.exists()
    assert test_script_2.exists()


@pytest.mark.slow_test
def test_ssh_run_pre_flight_args_prevent_injection(
    salt_ssh_cli, _create_roster, tmp_path
):
    """
    test ssh when --pre-flight is passed to salt-ssh
    and evil arguments are used in order to produce shell injection
    """
    injected_file = tmp_path / "injection"
    _custom_roster(
        salt_ssh_cli.roster_file,
        {"ssh_pre_flight_args": f"foobar; echo injected > {str(injected_file)}"},
    )
    # Create pre_flight script that accepts args
    test_script = _create_roster["test_script"]
    test_script_1 = pathlib.Path(test_script + "-echo")
    test_script_2 = pathlib.Path(test_script + "-foobar;")
    with salt.utils.files.fopen(_create_roster["data"]["ssh_pre_flight"], "w") as fp:
        fp.write(
            f"""
        touch {str(test_script)}-$1
        touch {str(test_script)}-$2
        """
        )

    # make sure we previously ran a command so the thin dir exists
    ret = salt_ssh_cli.run("test.ping")
    assert ret.returncode == 0
    assert test_script_1.exists()
    assert test_script_2.exists()
    test_script_1.unlink()
    test_script_2.unlink()
    assert not injected_file.is_file()

    ret = salt_ssh_cli.run("test.ping", "--pre-flight")
    assert ret.returncode == 0

    assert test_script_1.exists()
    assert test_script_2.exists()
    assert (
        not injected_file.is_file()
    ), "File injection suceeded. This shouldn't happend"


@pytest.mark.flaky(max_runs=4)
@pytest.mark.slow_test
def test_ssh_run_pre_flight_failure(salt_ssh_cli, _create_roster):
    """
    test ssh_pre_flight when there is a failure
    in the script.
    """
    with salt.utils.files.fopen(_create_roster["data"]["ssh_pre_flight"], "w") as fp_:
        fp_.write("exit 2")

    ret = salt_ssh_cli.run("test.ping", "--pre-flight")
    assert ret.data["retcode"] == 2


@pytest.fixture
def account():
    username = random_string("test-account-", uppercase=False)
    with pytest.helpers.create_account(username=username) as account:
        yield account


@pytest.mark.slow_test
def test_ssh_pre_flight_script(salt_ssh_cli, caplog, _create_roster, tmp_path, account):
    """
    Test to ensure user cannot create and run a script
    with the expected pre_flight script path on target.
    """
    try:
        script = pathlib.Path.home() / "hacked"
        tmp_preflight = pathlib.Path("/tmp", "ssh_pre_flight.sh")
        tmp_preflight.write_text(f"touch {script}", encoding="utf-8")
        os.chown(tmp_preflight, account.info.uid, account.info.gid)
        ret = salt_ssh_cli.run("test.ping")
        assert not script.is_file()
        assert ret.returncode == 0
        assert ret.stdout == '{\n"localhost": true\n}\n'
    finally:
        for _file in [script, tmp_preflight]:
            if _file.is_file():
                _file.unlink()


def demote(user_uid, user_gid):
    def result():
        # os.setgid does not remove group membership, so we remove them here so they are REALLY non-root
        os.setgroups([])
        os.setgid(user_gid)
        os.setuid(user_uid)

    return result


@pytest.mark.slow_test
def test_ssh_pre_flight_perms(salt_ssh_cli, caplog, _create_roster, account):
    """
    Test to ensure standard user cannot run pre flight script
    on target when user sets wrong permissions (777) on
    ssh_pre_flight script.
    """
    try:
        script = pathlib.Path("/tmp", "itworked")
        preflight = pathlib.Path("/ssh_pre_flight.sh")
        preflight.write_text(f"touch {str(script)}", encoding="utf-8")
        tmp_preflight = pathlib.Path("/tmp", preflight.name)

        _custom_roster(salt_ssh_cli.roster_file, {"ssh_pre_flight": str(preflight)})
        preflight.chmod(0o0777)
        run_script = pathlib.Path("/run_script")
        run_script.write_text(
            f"""
        x=1
        while [ $x -le 200000 ]; do
            SCRIPT=`bash {str(tmp_preflight)} 2> /dev/null; echo $?`
            if [ ${{SCRIPT}} -eq 0 ]; then
                break
            fi
            x=$(( $x + 1 ))
        done
        """,
            encoding="utf-8",
        )
        run_script.chmod(0o0777)
        # pylint: disable=W1509
        ret = subprocess.Popen(
            ["sh", f"{run_script}"],
            preexec_fn=demote(account.info.uid, account.info.gid),
            stdout=None,
            stderr=None,
            stdin=None,
            universal_newlines=True,
        )
        # pylint: enable=W1509
        ret = salt_ssh_cli.run("test.ping")
        assert ret.returncode == 0

        # Lets make sure a different user other than root
        # Didn't run the script
        assert os.stat(script).st_uid != account.info.uid
        assert script.is_file()
    finally:
        for _file in [script, preflight, tmp_preflight, run_script]:
            if _file.is_file():
                _file.unlink()


@pytest.mark.slow_test
def test_ssh_run_pre_flight_target_file_perms(salt_ssh_cli, _create_roster, tmp_path):
    """
    test ssh_pre_flight to ensure the target pre flight script
    has the correct perms
    """
    perms_file = tmp_path / "perms"
    with salt.utils.files.fopen(_create_roster["data"]["ssh_pre_flight"], "w") as fp_:
        fp_.write(
            f"""
        SCRIPT_NAME=$0
        stat -L -c "%a %G %U" $SCRIPT_NAME > {perms_file}
        """
        )

    ret = salt_ssh_cli.run("test.ping", "--pre-flight")
    assert ret.returncode == 0
    with salt.utils.files.fopen(perms_file) as fp:
        data = fp.read()
    assert data.split()[0] == "600"
    uid = os.getuid()
    gid = os.getgid()
    assert data.split()[1] == grp.getgrgid(gid).gr_name
    assert data.split()[2] == pwd.getpwuid(uid).pw_name



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import errno
import os

import pytest

import salt.modules.puppet as puppet
import salt.utils.args
import salt.utils.files
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {puppet: {}}


@pytest.fixture
def puppet_config():
    _puppet_config = """
---
confdir: "/etc/puppet"
rundir: "/var/run/puppetlabs"
vardir: "/var/lib/puppet"
"""

    yield _puppet_config


def test_run(puppet_config):
    """
    Test to execute a puppet run
    """
    mock_empty_lst = []

    mock = MagicMock(return_value={"A": "B"})
    with patch.object(salt.utils.args, "clean_kwargs", mock):
        cmd_run_all_mock = MagicMock(return_value={"retcode": 0})
        cmd_run_mock = MagicMock(side_effect=[puppet_config, mock_empty_lst])
        with patch.dict(
            puppet.__salt__, {"cmd.run_all": cmd_run_all_mock, "cmd.run": cmd_run_mock}
        ):
            ret = puppet.run()
            assert ret


def test_noop():
    """
    Test to execute a puppet noop run
    """
    mock = MagicMock(return_value={"stderr": "A", "stdout": "B"})
    with patch.object(puppet, "run", mock):
        assert puppet.noop() == {"stderr": "A", "stdout": "B"}


def test_enable(puppet_config):
    """
    Test to enable the puppet agent
    """
    mock_empty_lst = []
    cmd_run_mock = MagicMock(
        side_effect=[puppet_config, puppet_config, puppet_config, mock_empty_lst]
    )
    with patch.dict(puppet.__salt__, {"cmd.run": cmd_run_mock}):
        mock = MagicMock(return_value=True)
        with patch.object(os.path, "isfile", mock):
            mock = MagicMock(return_value=True)
            with patch.object(os, "remove", mock):
                assert puppet.enable()
            with patch.object(os, "remove", MagicMock(side_effect=IOError)):
                pytest.raises(CommandExecutionError, puppet.enable)

        assert not puppet.enable()


def test_disable(puppet_config):
    """
    Test to disable the puppet agent
    """
    cmd_run_mock = MagicMock(return_value=puppet_config)
    with patch.dict(puppet.__salt__, {"cmd.run": cmd_run_mock}):
        mock = MagicMock(side_effect=[True, False])
        with patch.object(os.path, "isfile", mock):
            assert not puppet.disable()

            with patch("salt.utils.files.fopen", mock_open()):
                assert puppet.disable()

            try:
                with patch("salt.utils.files.fopen", mock_open()) as m_open:
                    m_open.side_effect = IOError(13, "Permission denied:", "/file")
                    pytest.raises(CommandExecutionError, puppet.disable)
            except StopIteration:
                pass


def test_status(puppet_config):
    """
    Test to display puppet agent status
    """
    cmd_run_mock = MagicMock(return_value=puppet_config)
    with patch.dict(puppet.__salt__, {"cmd.run": cmd_run_mock}):
        mock = MagicMock(side_effect=[True])
        with patch.object(os.path, "isfile", mock):
            assert puppet.status() == "Administratively disabled"

        mock = MagicMock(side_effect=[False, True])
        with patch.object(os.path, "isfile", mock):
            with patch("salt.utils.files.fopen", mock_open(read_data="1")):
                mock = MagicMock(return_value=True)
                with patch.object(os, "kill", mock):
                    assert puppet.status() == "Applying a catalog"

        mock = MagicMock(side_effect=[False, True])
        with patch.object(os.path, "isfile", mock):
            with patch("salt.utils.files.fopen", mock_open()):
                mock = MagicMock(return_value=True)
                with patch.object(os, "kill", mock):
                    assert puppet.status() == "Stale lockfile"

        mock = MagicMock(side_effect=[False, False, True])
        with patch.object(os.path, "isfile", mock):
            with patch("salt.utils.files.fopen", mock_open(read_data="1")):
                mock = MagicMock(return_value=True)
                with patch.object(os, "kill", mock):
                    assert puppet.status() == "Idle daemon"

        mock = MagicMock(side_effect=[False, False, True])
        with patch.object(os.path, "isfile", mock):
            with patch("salt.utils.files.fopen", mock_open()):
                mock = MagicMock(return_value=True)
                with patch.object(os, "kill", mock):
                    assert puppet.status() == "Stale pidfile"

        mock = MagicMock(side_effect=[False, False, False])
        with patch.object(os.path, "isfile", mock):
            assert puppet.status() == "Stopped"


def test_summary(puppet_config):
    """
    Test to show a summary of the last puppet agent run
    """
    cmd_run_mock = MagicMock(return_value=puppet_config)
    with patch.dict(puppet.__salt__, {"cmd.run": cmd_run_mock}):
        with patch("salt.utils.files.fopen", mock_open(read_data="resources: 1")):
            assert puppet.summary() == {"resources": 1}

        permission_error = IOError(errno.EACCES, "Permission denied:", "/file")
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=permission_error)
        ) as m_open:
            pytest.raises(CommandExecutionError, puppet.summary)


def test_plugin_sync(puppet_config):
    """
    Test to runs a plugin synch between the puppet master and agent
    """
    cmd_run_mock = MagicMock(return_value=puppet_config)
    with patch.dict(puppet.__salt__, {"cmd.run": cmd_run_mock}):
        mock_lst = MagicMock(side_effect=[False, True])
        with patch.dict(puppet.__salt__, {"cmd.run": mock_lst}):
            assert puppet.plugin_sync() == ""

            assert puppet.plugin_sync()


def test_facts():
    """
    Test to run facter and return the results
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": "1\n2"})
    with patch.dict(puppet.__salt__, {"cmd.run_all": mock}):
        mock = MagicMock(side_effect=[["a", "b"], ["c", "d"]])
        with patch.object(puppet, "_format_fact", mock):
            assert puppet.facts() == {"a": "b", "c": "d"}


def test_fact():
    """
    Test to run facter for a specific fact
    """
    mock = MagicMock(
        side_effect=[
            {"retcode": 0, "stdout": False},
            {"retcode": 0, "stdout": True},
        ]
    )
    with patch.dict(puppet.__salt__, {"cmd.run_all": mock}):
        assert puppet.fact("salt") == ""

        assert puppet.fact("salt")



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.makeconf as makeconf
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {makeconf: {}}


def test_present():
    """
    Test to verify that the variable is in the ``make.conf``
    and has the provided settings.
    """
    name = "makeopts"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock_t = MagicMock(return_value=True)
    with patch.dict(makeconf.__salt__, {"makeconf.get_var": mock_t}):
        comt = f"Variable {name} is already present in make.conf"
        ret.update({"comment": comt})
        assert makeconf.present(name) == ret


def test_absent():
    """
    Test to verify that the variable is not in the ``make.conf``.
    """
    name = "makeopts"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock = MagicMock(return_value=None)
    with patch.dict(makeconf.__salt__, {"makeconf.get_var": mock}):
        comt = f"Variable {name} is already absent from make.conf"
        ret.update({"comment": comt})
        assert makeconf.absent(name) == ret



"""
Management of PostgreSQL groups (roles)
=======================================

The postgres_group module is used to create and manage Postgres groups.

.. code-block:: yaml

    frank:
      postgres_group.present
"""

import logging

# Salt imports
from salt.modules import postgres

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if the postgres module is present
    """
    if "postgres.group_create" not in __salt__:
        return (
            False,
            "Unable to load postgres module.  Make sure `postgres.bins_dir` is set.",
        )
    return True


def present(
    name,
    createdb=None,
    createroles=None,
    encrypted=None,
    superuser=None,
    inherit=None,
    login=None,
    replication=None,
    password=None,
    refresh_password=None,
    groups=None,
    user=None,
    maintenance_db=None,
    db_password=None,
    db_host=None,
    db_port=None,
    db_user=None,
):
    """
    Ensure that the named group is present with the specified privileges
    Please note that the user/group notion in postgresql is just abstract, we
    have roles, where users can be seen as roles with the ``LOGIN`` privilege
    and groups the others.

    name
        The name of the group to manage

    createdb
        Is the group allowed to create databases?

    createroles
        Is the group allowed to create other roles/users

    encrypted
        How the password should be stored.

        If encrypted is ``None``, ``True``, or ``md5``, it will use
        PostgreSQL's MD5 algorithm.

        If encrypted is ``False``, it will be stored in plaintext.

        If encrypted is ``scram-sha-256``, it will use the algorithm described
        in RFC 7677.

        .. versionchanged:: 3003

            Prior versions only supported ``True`` and ``False``

    login
        Should the group have login perm

    inherit
        Should the group inherit permissions

    superuser
        Should the new group be a "superuser"

    replication
        Should the new group be allowed to initiate streaming replication

    password
        The group's password.
        It can be either a plain string or a pre-hashed password::

            'md5{MD5OF({password}{role}}'
            'SCRAM-SHA-256${iterations}:{salt}${stored_key}:{server_key}'

        If encrypted is not ``False``, then the password will be converted
        to the appropriate format above, if not already. As a consequence,
        passwords that start with "md5" or "SCRAM-SHA-256" cannot be used.

    refresh_password
        Password refresh flag

        Boolean attribute to specify whether to password comparison check
        should be performed.

        If refresh_password is ``True``, the password will be automatically
        updated without extra password change check.

        This behaviour makes it possible to execute in environments without
        superuser access available, e.g. Amazon RDS for PostgreSQL

    groups
        A string of comma separated groups the group should be in

    user
        System user all operations should be performed on behalf of

        .. versionadded:: 0.17.0

    db_user
        database username if different from config or default

    db_password
        user password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Group {name} is already present",
    }

    # default to encrypted passwords
    if encrypted is None:
        encrypted = postgres._DEFAULT_PASSWORDS_ENCRYPTION

    db_args = {
        "maintenance_db": maintenance_db,
        "runas": user,
        "host": db_host,
        "user": db_user,
        "port": db_port,
        "password": db_password,
    }

    # check if group exists
    mode = "create"
    group_attr = __salt__["postgres.role_get"](
        name, return_password=not refresh_password, **db_args
    )
    if group_attr is not None:
        mode = "update"

    if password is not None:
        if (
            mode == "update"
            and not refresh_password
            and postgres._verify_password(
                name, password, group_attr["password"], encrypted
            )
        ):
            # if password already matches then don't touch it
            password = None
        else:
            # encrypt password if necessary
            password = postgres._maybe_encrypt_password(
                name, password, encrypted=encrypted
            )

    # The user is not present, make it!
    update = {}
    if mode == "update":
        role_groups = group_attr.get("groups", [])
        if createdb is not None and group_attr["can create databases"] != createdb:
            update["createdb"] = createdb
        if inherit is not None and group_attr["inherits privileges"] != inherit:
            update["inherit"] = inherit
        if login is not None and group_attr["can login"] != login:
            update["login"] = login
        if createroles is not None and group_attr["can create roles"] != createroles:
            update["createroles"] = createroles
        if replication is not None and group_attr["replication"] != replication:
            update["replication"] = replication
        if superuser is not None and group_attr["superuser"] != superuser:
            update["superuser"] = superuser
        if password is not None:
            update["password"] = True
        if groups is not None:
            lgroups = groups
            if isinstance(groups, str):
                lgroups = lgroups.split(",")
            if isinstance(lgroups, list):
                missing_groups = [a for a in lgroups if a not in role_groups]
                if missing_groups:
                    update["groups"] = missing_groups

    if mode == "create" or (mode == "update" and update):
        if __opts__["test"]:
            if update:
                ret["changes"][name] = update
            ret["result"] = None
            ret["comment"] = f"Group {name} is set to be {mode}d"
            return ret
        cret = __salt__[f"postgres.group_{mode}"](
            groupname=name,
            createdb=createdb,
            createroles=createroles,
            encrypted=encrypted,
            login=login,
            inherit=inherit,
            superuser=superuser,
            replication=replication,
            rolepassword=password,
            groups=groups,
            **db_args,
        )
    else:
        cret = None

    if cret:
        ret["comment"] = f"The group {name} has been {mode}d"
        if update:
            ret["changes"][name] = update
        else:
            ret["changes"][name] = "Present"
    elif cret is not None:
        ret["comment"] = f"Failed to {mode} group {name}"
        ret["result"] = False
    else:
        ret["result"] = True

    return ret


def absent(
    name,
    user=None,
    maintenance_db=None,
    db_password=None,
    db_host=None,
    db_port=None,
    db_user=None,
):
    """
    Ensure that the named group is absent

    name
        The groupname of the group to remove

    user
        System user all operations should be performed on behalf of

        .. versionadded:: 0.17.0

    db_user
        database username if different from config or default

    db_password
        user password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    db_args = {
        "maintenance_db": maintenance_db,
        "runas": user,
        "host": db_host,
        "user": db_user,
        "port": db_port,
        "password": db_password,
    }
    # check if group exists and remove it
    if __salt__["postgres.user_exists"](name, **db_args):
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Group {name} is set to be removed"
            return ret
        if __salt__["postgres.group_remove"](name, **db_args):
            ret["comment"] = f"Group {name} has been removed"
            ret["changes"][name] = "Absent"
            return ret
        else:
            ret["result"] = False
            ret["comment"] = f"Group {name} failed to be removed"
            return ret
    else:
        ret["comment"] = f"Group {name} is not present, so it cannot be removed"

    return ret



import pathlib
import sys


def setup(pth_file_path):
    # Discover the extras-<py-major>.<py-minor> directory
    extras_parent_path = pathlib.Path(pth_file_path).resolve().parent.parent
    if not sys.platform.startswith("win"):
        extras_parent_path = extras_parent_path.parent

    extras_path = str(extras_parent_path / "extras-{}.{}".format(*sys.version_info))

    if extras_path in sys.path and sys.path[0] != extras_path:
        # The extras directory must come first
        sys.path.remove(extras_path)

    if extras_path not in sys.path:
        sys.path.insert(0, extras_path)



"""
Azure (ARM) Network Execution Module

.. versionadded:: 2019.2.0

.. warning::

    This cloud provider will be removed from Salt in version 3007 in favor of
    the `saltext.azurerm Salt Extension
    <https://github.com/salt-extensions/saltext-azurerm>`_

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 2.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.8
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 1.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 1.0.0
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 1.7.1
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 1.1.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 1.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.32.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.4.21
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as keyword arguments
    to every function in order to work properly.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

**cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud.
    Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

"""

# Python libs

import logging
from functools import wraps

# Salt libs
import salt.utils.azurearm
from salt.exceptions import SaltInvocationError  # pylint: disable=unused-import

# Azure libs
HAS_LIBS = False
try:
    import azure.mgmt.network.models  # pylint: disable=unused-import
    from msrest.exceptions import SerializationError
    from msrestazure.azure_exceptions import CloudError

    HAS_LIBS = True
except ImportError:
    pass

__virtualname__ = "azurearm_network"

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_LIBS:
        return (
            False,
            "The following dependencies are required to use the AzureARM modules: "
            "Microsoft Azure SDK for Python >= 2.0rc6, "
            "MS REST Azure (msrestazure) >= 0.4",
        )

    return __virtualname__


def _deprecation_message(function):
    """
    Decorator wrapper to warn about azurearm deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            "Chlorine",
            "The 'azurearm' functionality in Salt has been deprecated and its "
            "functionality will be removed in version 3007 in favor of the "
            "saltext.azurerm Salt Extension. "
            "(https://github.com/salt-extensions/saltext-azurerm)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def check_dns_name_availability(name, region, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Check whether a domain name in the current zone is available for use.

    :param name: The DNS name to query.

    :param region: The region to query for the DNS name in question.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.check_dns_name_availability testdnsname westus

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        check_dns_name = netconn.check_dns_name_availability(
            location=region, domain_name_label=name
        )
        result = check_dns_name.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def check_ip_address_availability(
    ip_address, virtual_network, resource_group, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Check that a private ip address is available within the specified
    virtual network.

    :param ip_address: The ip_address to query.

    :param virtual_network: The virtual network to query for the IP address
        in question.

    :param resource_group: The resource group name assigned to the
        virtual network.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.check_ip_address_availability 10.0.0.4 testnet testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        check_ip = netconn.virtual_networks.check_ip_address_availability(
            resource_group_name=resource_group,
            virtual_network_name=virtual_network,
            ip_address=ip_address,
        )
        result = check_ip.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def default_security_rule_get(name, security_group, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a default security rule within a security group.

    :param name: The name of the security rule to query.

    :param security_group: The network security group containing the
        security rule.

    :param resource_group: The resource group name assigned to the
        network security group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.default_security_rule_get DenyAllOutBound testnsg testgroup

    """
    result = {}

    default_rules = default_security_rules_list(
        security_group=security_group, resource_group=resource_group, **kwargs
    )

    if isinstance(default_rules, dict) and "error" in default_rules:
        return default_rules

    try:
        for default_rule in default_rules:
            if default_rule["name"] == name:
                result = default_rule
        if not result:
            result = {"error": f"Unable to find {name} in {security_group}!"}
    except KeyError as exc:
        log.error("Unable to find %s in %s!", name, security_group)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def default_security_rules_list(security_group, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List default security rules within a security group.

    :param security_group: The network security group to query.

    :param resource_group: The resource group name assigned to the
        network security group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.default_security_rules_list testnsg testgroup

    """
    result = {}

    secgroup = network_security_group_get(
        security_group=security_group, resource_group=resource_group, **kwargs
    )

    if "error" in secgroup:
        return secgroup

    try:
        result = secgroup["default_security_rules"]
    except KeyError as exc:
        log.error("No default security rules found for %s!", security_group)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def security_rules_list(security_group, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List security rules within a network security group.

    :param security_group: The network security group to query.

    :param resource_group: The resource group name assigned to the
        network security group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.security_rules_list testnsg testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        secrules = netconn.security_rules.list(
            network_security_group_name=security_group,
            resource_group_name=resource_group,
        )
        result = __utils__["azurearm.paged_object_to_list"](secrules)
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def security_rule_create_or_update(
    name,
    access,
    direction,
    priority,
    protocol,
    security_group,
    resource_group,
    source_address_prefix=None,
    destination_address_prefix=None,
    source_port_range=None,
    destination_port_range=None,
    source_address_prefixes=None,
    destination_address_prefixes=None,
    source_port_ranges=None,
    destination_port_ranges=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Create or update a security rule within a specified network security group.

    :param name: The name of the security rule to create.

    :param access:
        'allow' or 'deny'

    :param direction:
        'inbound' or 'outbound'

    :param priority:
        Integer between 100 and 4096 used for ordering rule application.

    :param protocol:
        'tcp', 'udp', or '*'

    :param destination_address_prefix:
        The CIDR or destination IP range. Asterix '*' can also be used to match all destination IPs.
        Default tags such as 'VirtualNetwork', 'AzureLoadBalancer' and 'Internet' can also be used.
        If this is an ingress rule, specifies where network traffic originates from.

    :param destination_port_range:
        The destination port or range. Integer or range between 0 and 65535. Asterix '*'
        can also be used to match all ports.

    :param source_address_prefix:
        The CIDR or source IP range. Asterix '*' can also be used to match all source IPs.
        Default tags such as 'VirtualNetwork', 'AzureLoadBalancer' and 'Internet' can also be used.
        If this is an ingress rule, specifies where network traffic originates from.

    :param source_port_range:
        The source port or range. Integer or range between 0 and 65535. Asterix '*'
        can also be used to match all ports.

    :param destination_address_prefixes:
        A list of destination_address_prefix values. This parameter overrides destination_address_prefix
        and will cause any value entered there to be ignored.

    :param destination_port_ranges:
        A list of destination_port_range values. This parameter overrides destination_port_range
        and will cause any value entered there to be ignored.

    :param source_address_prefixes:
        A list of source_address_prefix values. This parameter overrides source_address_prefix
        and will cause any value entered there to be ignored.

    :param source_port_ranges:
        A list of source_port_range values. This parameter overrides source_port_range
        and will cause any value entered there to be ignored.

    :param security_group: The network security group containing the
        security rule.

    :param resource_group: The resource group name assigned to the
        network security group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.security_rule_create_or_update testrule1 allow outbound 101 tcp testnsg testgroup \
                  source_address_prefix='*' destination_address_prefix=internet source_port_range='*' \
                  destination_port_range='1-1024'

    """
    exclusive_params = [
        ("source_port_ranges", "source_port_range"),
        ("source_address_prefixes", "source_address_prefix"),
        ("destination_port_ranges", "destination_port_range"),
        ("destination_address_prefixes", "destination_address_prefix"),
    ]

    for params in exclusive_params:
        # pylint: disable=eval-used
        if not eval(params[0]) and not eval(params[1]):
            log.error(
                "Either the %s or %s parameter must be provided!", params[0], params[1]
            )
            return False
        # pylint: disable=eval-used
        if eval(params[0]):
            # pylint: disable=exec-used
            exec(f"{params[1]} = None")

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        rulemodel = __utils__["azurearm.create_object_model"](
            "network",
            "SecurityRule",
            name=name,
            access=access,
            direction=direction,
            priority=priority,
            protocol=protocol,
            source_port_ranges=source_port_ranges,
            source_port_range=source_port_range,
            source_address_prefixes=source_address_prefixes,
            source_address_prefix=source_address_prefix,
            destination_port_ranges=destination_port_ranges,
            destination_port_range=destination_port_range,
            destination_address_prefixes=destination_address_prefixes,
            destination_address_prefix=destination_address_prefix,
            **kwargs,
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        secrule = netconn.security_rules.create_or_update(
            resource_group_name=resource_group,
            network_security_group_name=security_group,
            security_rule_name=name,
            security_rule_parameters=rulemodel,
        )
        secrule.wait()
        secrule_result = secrule.result()
        result = secrule_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def security_rule_delete(security_rule, security_group, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a security rule within a specified security group.

    :param name: The name of the security rule to delete.

    :param security_group: The network security group containing the
        security rule.

    :param resource_group: The resource group name assigned to the
        network security group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.security_rule_delete testrule1 testnsg testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        secrule = netconn.security_rules.delete(
            network_security_group_name=security_group,
            resource_group_name=resource_group,
            security_rule_name=security_rule,
        )
        secrule.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def security_rule_get(security_rule, security_group, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get a security rule within a specified network security group.

    :param name: The name of the security rule to query.

    :param security_group: The network security group containing the
        security rule.

    :param resource_group: The resource group name assigned to the
        network security group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.security_rule_get testrule1 testnsg testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        secrule = netconn.security_rules.get(
            network_security_group_name=security_group,
            resource_group_name=resource_group,
            security_rule_name=security_rule,
        )
        result = secrule.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def network_security_group_create_or_update(
    name, resource_group, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 2019.2.0

    Create or update a network security group.

    :param name: The name of the network security group to create.

    :param resource_group: The resource group name assigned to the
        network security group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_security_group_create_or_update testnsg testgroup

    """
    if "location" not in kwargs:
        rg_props = __salt__["azurearm_resource.resource_group_get"](
            resource_group, **kwargs
        )

        if "error" in rg_props:
            log.error("Unable to determine location from resource group specified.")
            return False
        kwargs["location"] = rg_props["location"]

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        secgroupmodel = __utils__["azurearm.create_object_model"](
            "network", "NetworkSecurityGroup", **kwargs
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        secgroup = netconn.network_security_groups.create_or_update(
            resource_group_name=resource_group,
            network_security_group_name=name,
            parameters=secgroupmodel,
        )
        secgroup.wait()
        secgroup_result = secgroup.result()
        result = secgroup_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def network_security_group_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a network security group within a resource group.

    :param name: The name of the network security group to delete.

    :param resource_group: The resource group name assigned to the
        network security group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_security_group_delete testnsg testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        secgroup = netconn.network_security_groups.delete(
            resource_group_name=resource_group, network_security_group_name=name
        )
        secgroup.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def network_security_group_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a network security group within a resource group.

    :param name: The name of the network security group to query.

    :param resource_group: The resource group name assigned to the
        network security group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_security_group_get testnsg testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        secgroup = netconn.network_security_groups.get(
            resource_group_name=resource_group, network_security_group_name=name
        )
        result = secgroup.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def network_security_groups_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all network security groups within a resource group.

    :param resource_group: The resource group name to list network security \
        groups within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_security_groups_list testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        secgroups = __utils__["azurearm.paged_object_to_list"](
            netconn.network_security_groups.list(resource_group_name=resource_group)
        )
        for secgroup in secgroups:
            result[secgroup["name"]] = secgroup
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def network_security_groups_list_all(**kwargs):  # pylint: disable=invalid-name
    """
    .. versionadded:: 2019.2.0

    List all network security groups within a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_security_groups_list_all

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        secgroups = __utils__["azurearm.paged_object_to_list"](
            netconn.network_security_groups.list_all()
        )
        for secgroup in secgroups:
            result[secgroup["name"]] = secgroup
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def subnets_list(virtual_network, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all subnets within a virtual network.

    :param virtual_network: The virtual network name to list subnets within.

    :param resource_group: The resource group name assigned to the
        virtual network.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.subnets_list testnet testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        subnets = __utils__["azurearm.paged_object_to_list"](
            netconn.subnets.list(
                resource_group_name=resource_group, virtual_network_name=virtual_network
            )
        )

        for subnet in subnets:
            result[subnet["name"]] = subnet
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def subnet_get(name, virtual_network, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific subnet.

    :param name: The name of the subnet to query.

    :param virtual_network: The virtual network name containing the
        subnet.

    :param resource_group: The resource group name assigned to the
        virtual network.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.subnet_get testsubnet testnet testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        subnet = netconn.subnets.get(
            resource_group_name=resource_group,
            virtual_network_name=virtual_network,
            subnet_name=name,
        )

        result = subnet.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def subnet_create_or_update(
    name, address_prefix, virtual_network, resource_group, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Create or update a subnet.

    :param name: The name assigned to the subnet being created or updated.

    :param address_prefix: A valid CIDR block within the virtual network.

    :param virtual_network: The virtual network name containing the
        subnet.

    :param resource_group: The resource group name assigned to the
        virtual network.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.subnet_create_or_update testsubnet \
                  '10.0.0.0/24' testnet testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    # Use NSG name to link to the ID of an existing NSG.
    if kwargs.get("network_security_group"):
        nsg = network_security_group_get(
            name=kwargs["network_security_group"],
            resource_group=resource_group,
            **kwargs,
        )
        if "error" not in nsg:
            kwargs["network_security_group"] = {"id": str(nsg["id"])}

    # Use Route Table name to link to the ID of an existing Route Table.
    if kwargs.get("route_table"):
        rt_table = route_table_get(
            name=kwargs["route_table"], resource_group=resource_group, **kwargs
        )
        if "error" not in rt_table:
            kwargs["route_table"] = {"id": str(rt_table["id"])}

    try:
        snetmodel = __utils__["azurearm.create_object_model"](
            "network",
            "Subnet",
            address_prefix=address_prefix,
            resource_group=resource_group,
            **kwargs,
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        subnet = netconn.subnets.create_or_update(
            resource_group_name=resource_group,
            virtual_network_name=virtual_network,
            subnet_name=name,
            subnet_parameters=snetmodel,
        )
        subnet.wait()
        sn_result = subnet.result()
        result = sn_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def subnet_delete(name, virtual_network, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a subnet.

    :param name: The name of the subnet to delete.

    :param virtual_network: The virtual network name containing the
        subnet.

    :param resource_group: The resource group name assigned to the
        virtual network.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.subnet_delete testsubnet testnet testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        subnet = netconn.subnets.delete(
            resource_group_name=resource_group,
            virtual_network_name=virtual_network,
            subnet_name=name,
        )
        subnet.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def virtual_networks_list_all(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all virtual networks within a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.virtual_networks_list_all

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        vnets = __utils__["azurearm.paged_object_to_list"](
            netconn.virtual_networks.list_all()
        )

        for vnet in vnets:
            result[vnet["name"]] = vnet
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_networks_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all virtual networks within a resource group.

    :param resource_group: The resource group name to list virtual networks
        within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.virtual_networks_list testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        vnets = __utils__["azurearm.paged_object_to_list"](
            netconn.virtual_networks.list(resource_group_name=resource_group)
        )

        for vnet in vnets:
            result[vnet["name"]] = vnet
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


# pylint: disable=invalid-name
@_deprecation_message
def virtual_network_create_or_update(name, address_prefixes, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Create or update a virtual network.

    :param name: The name assigned to the virtual network being
        created or updated.

    :param address_prefixes: A list of CIDR blocks which can be used
        by subnets within the virtual network.

    :param resource_group: The resource group name assigned to the
        virtual network.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.virtual_network_create_or_update \
                  testnet ['10.0.0.0/16'] testgroup

    """
    if "location" not in kwargs:
        rg_props = __salt__["azurearm_resource.resource_group_get"](
            resource_group, **kwargs
        )

        if "error" in rg_props:
            log.error("Unable to determine location from resource group specified.")
            return False
        kwargs["location"] = rg_props["location"]

    if not isinstance(address_prefixes, list):
        log.error("Address prefixes must be specified as a list!")
        return False

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    address_space = {"address_prefixes": address_prefixes}
    dhcp_options = {"dns_servers": kwargs.get("dns_servers")}

    try:
        vnetmodel = __utils__["azurearm.create_object_model"](
            "network",
            "VirtualNetwork",
            address_space=address_space,
            dhcp_options=dhcp_options,
            **kwargs,
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        vnet = netconn.virtual_networks.create_or_update(
            virtual_network_name=name,
            resource_group_name=resource_group,
            parameters=vnetmodel,
        )
        vnet.wait()
        vnet_result = vnet.result()
        result = vnet_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def virtual_network_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a virtual network.

    :param name: The name of the virtual network to delete.

    :param resource_group: The resource group name assigned to the
        virtual network

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.virtual_network_delete testnet testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        vnet = netconn.virtual_networks.delete(
            virtual_network_name=name, resource_group_name=resource_group
        )
        vnet.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def virtual_network_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific virtual network.

    :param name: The name of the virtual network to query.

    :param resource_group: The resource group name assigned to the
        virtual network.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.virtual_network_get testnet testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        vnet = netconn.virtual_networks.get(
            virtual_network_name=name, resource_group_name=resource_group
        )
        result = vnet.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def load_balancers_list_all(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all load balancers within a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.load_balancers_list_all

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        load_balancers = __utils__["azurearm.paged_object_to_list"](
            netconn.load_balancers.list_all()
        )

        for load_balancer in load_balancers:
            result[load_balancer["name"]] = load_balancer
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def load_balancers_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all load balancers within a resource group.

    :param resource_group: The resource group name to list load balancers
        within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.load_balancers_list testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        load_balancers = __utils__["azurearm.paged_object_to_list"](
            netconn.load_balancers.list(resource_group_name=resource_group)
        )

        for load_balancer in load_balancers:
            result[load_balancer["name"]] = load_balancer
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def load_balancer_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific load balancer.

    :param name: The name of the load balancer to query.

    :param resource_group: The resource group name assigned to the
        load balancer.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.load_balancer_get testlb testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        load_balancer = netconn.load_balancers.get(
            load_balancer_name=name, resource_group_name=resource_group
        )
        result = load_balancer.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def load_balancer_create_or_update(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Create or update a load balancer within a specified resource group.

    :param name: The name of the load balancer to create.

    :param resource_group: The resource group name assigned to the
        load balancer.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.load_balancer_create_or_update testlb testgroup

    """
    if "location" not in kwargs:
        rg_props = __salt__["azurearm_resource.resource_group_get"](
            resource_group, **kwargs
        )

        if "error" in rg_props:
            log.error("Unable to determine location from resource group specified.")
            return False
        kwargs["location"] = rg_props["location"]

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    if isinstance(kwargs.get("frontend_ip_configurations"), list):
        for idx in range(0, len(kwargs["frontend_ip_configurations"])):
            # Use Public IP Address name to link to the ID of an existing Public IP
            if "public_ip_address" in kwargs["frontend_ip_configurations"][idx]:
                pub_ip = public_ip_address_get(
                    name=kwargs["frontend_ip_configurations"][idx]["public_ip_address"],
                    resource_group=resource_group,
                    **kwargs,
                )
                if "error" not in pub_ip:
                    kwargs["frontend_ip_configurations"][idx]["public_ip_address"] = {
                        "id": str(pub_ip["id"])
                    }
            # Use Subnet name to link to the ID of an existing Subnet
            elif "subnet" in kwargs["frontend_ip_configurations"][idx]:
                vnets = virtual_networks_list(resource_group=resource_group, **kwargs)
                if "error" not in vnets:
                    for vnet in vnets:
                        subnets = subnets_list(
                            virtual_network=vnet,
                            resource_group=resource_group,
                            **kwargs,
                        )
                        if (
                            kwargs["frontend_ip_configurations"][idx]["subnet"]
                            in subnets
                        ):
                            kwargs["frontend_ip_configurations"][idx]["subnet"] = {
                                "id": str(
                                    subnets[
                                        kwargs["frontend_ip_configurations"][idx][
                                            "subnet"
                                        ]
                                    ]["id"]
                                )
                            }
                            break

    id_url = "/subscriptions/{0}/resourceGroups/{1}/providers/Microsoft.Network/loadBalancers/{2}/{3}/{4}"

    if isinstance(kwargs.get("load_balancing_rules"), list):
        for idx in range(0, len(kwargs["load_balancing_rules"])):
            # Link to sub-objects which might be created at the same time as the load balancer
            if "frontend_ip_configuration" in kwargs["load_balancing_rules"][idx]:
                kwargs["load_balancing_rules"][idx]["frontend_ip_configuration"] = {
                    "id": id_url.format(
                        kwargs.get("subscription_id"),
                        resource_group,
                        name,
                        "frontendIPConfigurations",
                        kwargs["load_balancing_rules"][idx][
                            "frontend_ip_configuration"
                        ],
                    )
                }
            if "backend_address_pool" in kwargs["load_balancing_rules"][idx]:
                kwargs["load_balancing_rules"][idx]["backend_address_pool"] = {
                    "id": id_url.format(
                        kwargs.get("subscription_id"),
                        resource_group,
                        name,
                        "backendAddressPools",
                        kwargs["load_balancing_rules"][idx]["backend_address_pool"],
                    )
                }
            if "probe" in kwargs["load_balancing_rules"][idx]:
                kwargs["load_balancing_rules"][idx]["probe"] = {
                    "id": id_url.format(
                        kwargs.get("subscription_id"),
                        resource_group,
                        name,
                        "probes",
                        kwargs["load_balancing_rules"][idx]["probe"],
                    )
                }

    if isinstance(kwargs.get("inbound_nat_rules"), list):
        for idx in range(0, len(kwargs["inbound_nat_rules"])):
            # Link to sub-objects which might be created at the same time as the load balancer
            if "frontend_ip_configuration" in kwargs["inbound_nat_rules"][idx]:
                kwargs["inbound_nat_rules"][idx]["frontend_ip_configuration"] = {
                    "id": id_url.format(
                        kwargs.get("subscription_id"),
                        resource_group,
                        name,
                        "frontendIPConfigurations",
                        kwargs["inbound_nat_rules"][idx]["frontend_ip_configuration"],
                    )
                }

    if isinstance(kwargs.get("inbound_nat_pools"), list):
        for idx in range(0, len(kwargs["inbound_nat_pools"])):
            # Link to sub-objects which might be created at the same time as the load balancer
            if "frontend_ip_configuration" in kwargs["inbound_nat_pools"][idx]:
                kwargs["inbound_nat_pools"][idx]["frontend_ip_configuration"] = {
                    "id": id_url.format(
                        kwargs.get("subscription_id"),
                        resource_group,
                        name,
                        "frontendIPConfigurations",
                        kwargs["inbound_nat_pools"][idx]["frontend_ip_configuration"],
                    )
                }

    if isinstance(kwargs.get("outbound_nat_rules"), list):
        for idx in range(0, len(kwargs["outbound_nat_rules"])):
            # Link to sub-objects which might be created at the same time as the load balancer
            if "frontend_ip_configuration" in kwargs["outbound_nat_rules"][idx]:
                kwargs["outbound_nat_rules"][idx]["frontend_ip_configuration"] = {
                    "id": id_url.format(
                        kwargs.get("subscription_id"),
                        resource_group,
                        name,
                        "frontendIPConfigurations",
                        kwargs["outbound_nat_rules"][idx]["frontend_ip_configuration"],
                    )
                }
            if "backend_address_pool" in kwargs["outbound_nat_rules"][idx]:
                kwargs["outbound_nat_rules"][idx]["backend_address_pool"] = {
                    "id": id_url.format(
                        kwargs.get("subscription_id"),
                        resource_group,
                        name,
                        "backendAddressPools",
                        kwargs["outbound_nat_rules"][idx]["backend_address_pool"],
                    )
                }

    try:
        lbmodel = __utils__["azurearm.create_object_model"](
            "network", "LoadBalancer", **kwargs
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        load_balancer = netconn.load_balancers.create_or_update(
            resource_group_name=resource_group,
            load_balancer_name=name,
            parameters=lbmodel,
        )
        load_balancer.wait()
        lb_result = load_balancer.result()
        result = lb_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def load_balancer_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a load balancer.

    :param name: The name of the load balancer to delete.

    :param resource_group: The resource group name assigned to the
        load balancer.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.load_balancer_delete testlb testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        load_balancer = netconn.load_balancers.delete(
            load_balancer_name=name, resource_group_name=resource_group
        )
        load_balancer.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def usages_list(location, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List subscription network usage for a location.

    :param location: The Azure location to query for network usage.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.usages_list westus

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        result = __utils__["azurearm.paged_object_to_list"](
            netconn.usages.list(location)
        )
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def network_interface_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a network interface.

    :param name: The name of the network interface to delete.

    :param resource_group: The resource group name assigned to the
        network interface.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_interface_delete test-iface0 testgroup

    """
    result = False

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        nic = netconn.network_interfaces.delete(
            network_interface_name=name, resource_group_name=resource_group
        )
        nic.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def network_interface_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific network interface.

    :param name: The name of the network interface to query.

    :param resource_group: The resource group name assigned to the
        network interface.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_interface_get test-iface0 testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        nic = netconn.network_interfaces.get(
            network_interface_name=name, resource_group_name=resource_group
        )
        result = nic.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


# pylint: disable=invalid-name
@_deprecation_message
def network_interface_create_or_update(
    name, ip_configurations, subnet, virtual_network, resource_group, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Create or update a network interface within a specified resource group.

    :param name: The name of the network interface to create.

    :param ip_configurations: A list of dictionaries representing valid
        NetworkInterfaceIPConfiguration objects. The 'name' key is required at
        minimum. At least one IP Configuration must be present.

    :param subnet: The name of the subnet assigned to the network interface.

    :param virtual_network: The name of the virtual network assigned to the subnet.

    :param resource_group: The resource group name assigned to the
        virtual network.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_interface_create_or_update test-iface0 [{'name': 'testipconfig1'}] \
                  testsubnet testnet testgroup

    """
    if "location" not in kwargs:
        rg_props = __salt__["azurearm_resource.resource_group_get"](
            resource_group, **kwargs
        )

        if "error" in rg_props:
            log.error("Unable to determine location from resource group specified.")
            return False
        kwargs["location"] = rg_props["location"]

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    # Use NSG name to link to the ID of an existing NSG.
    if kwargs.get("network_security_group"):
        nsg = network_security_group_get(
            name=kwargs["network_security_group"],
            resource_group=resource_group,
            **kwargs,
        )
        if "error" not in nsg:
            kwargs["network_security_group"] = {"id": str(nsg["id"])}

    # Use VM name to link to the ID of an existing VM.
    if kwargs.get("virtual_machine"):
        vm_instance = __salt__["azurearm_compute.virtual_machine_get"](
            name=kwargs["virtual_machine"], resource_group=resource_group, **kwargs
        )
        if "error" not in vm_instance:
            kwargs["virtual_machine"] = {"id": str(vm_instance["id"])}

    # Loop through IP Configurations and build each dictionary to pass to model creation.
    if isinstance(ip_configurations, list):
        subnet = subnet_get(
            name=subnet,
            virtual_network=virtual_network,
            resource_group=resource_group,
            **kwargs,
        )
        if "error" not in subnet:
            subnet = {"id": str(subnet["id"])}
            for ipconfig in ip_configurations:
                if "name" in ipconfig:
                    ipconfig["subnet"] = subnet
                    if isinstance(
                        ipconfig.get("application_gateway_backend_address_pools"), list
                    ):
                        # TODO: Add ID lookup for referenced object names
                        pass
                    if isinstance(
                        ipconfig.get("load_balancer_backend_address_pools"), list
                    ):
                        # TODO: Add ID lookup for referenced object names
                        pass
                    if isinstance(
                        ipconfig.get("load_balancer_inbound_nat_rules"), list
                    ):
                        # TODO: Add ID lookup for referenced object names
                        pass
                    if ipconfig.get("public_ip_address"):
                        pub_ip = public_ip_address_get(
                            name=ipconfig["public_ip_address"],
                            resource_group=resource_group,
                            **kwargs,
                        )
                        if "error" not in pub_ip:
                            ipconfig["public_ip_address"] = {"id": str(pub_ip["id"])}

    try:
        nicmodel = __utils__["azurearm.create_object_model"](
            "network", "NetworkInterface", ip_configurations=ip_configurations, **kwargs
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        interface = netconn.network_interfaces.create_or_update(
            resource_group_name=resource_group,
            network_interface_name=name,
            parameters=nicmodel,
        )
        interface.wait()
        nic_result = interface.result()
        result = nic_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def network_interfaces_list_all(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all network interfaces within a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_interfaces_list_all

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        nics = __utils__["azurearm.paged_object_to_list"](
            netconn.network_interfaces.list_all()
        )

        for nic in nics:
            result[nic["name"]] = nic
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def network_interfaces_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all network interfaces within a resource group.

    :param resource_group: The resource group name to list network
        interfaces within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_interfaces_list testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        nics = __utils__["azurearm.paged_object_to_list"](
            netconn.network_interfaces.list(resource_group_name=resource_group)
        )

        for nic in nics:
            result[nic["name"]] = nic
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


# pylint: disable=invalid-name
@_deprecation_message
def network_interface_get_effective_route_table(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get all route tables for a specific network interface.

    :param name: The name of the network interface to query.

    :param resource_group: The resource group name assigned to the
        network interface.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_interface_get_effective_route_table test-iface0 testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        nic = netconn.network_interfaces.get_effective_route_table(
            network_interface_name=name, resource_group_name=resource_group
        )
        nic.wait()
        tables = nic.result()
        tables = tables.as_dict()
        result = tables["value"]
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


# pylint: disable=invalid-name
@_deprecation_message
def network_interface_list_effective_network_security_groups(
    name, resource_group, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Get all network security groups applied to a specific network interface.

    :param name: The name of the network interface to query.

    :param resource_group: The resource group name assigned to the
        network interface.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.network_interface_list_effective_network_security_groups test-iface0 testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        nic = netconn.network_interfaces.list_effective_network_security_groups(
            network_interface_name=name, resource_group_name=resource_group
        )
        nic.wait()
        groups = nic.result()
        groups = groups.as_dict()
        result = groups["value"]
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


# pylint: disable=invalid-name
@_deprecation_message
def list_virtual_machine_scale_set_vm_network_interfaces(
    scale_set, vm_index, resource_group, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Get information about all network interfaces in a specific virtual machine within a scale set.

    :param scale_set: The name of the scale set to query.

    :param vm_index: The virtual machine index.

    :param resource_group: The resource group name assigned to the
        scale set.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.list_virtual_machine_scale_set_vm_network_interfaces testset testvm testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        nics = __utils__["azurearm.paged_object_to_list"](
            netconn.network_interfaces.list_virtual_machine_scale_set_vm_network_interfaces(
                virtual_machine_scale_set_name=scale_set,
                virtualmachine_index=vm_index,
                resource_group_name=resource_group,
            )
        )

        for nic in nics:
            result[nic["name"]] = nic
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


# pylint: disable=invalid-name
@_deprecation_message
def list_virtual_machine_scale_set_network_interfaces(
    scale_set, resource_group, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Get information about all network interfaces within a scale set.

    :param scale_set: The name of the scale set to query.

    :param resource_group: The resource group name assigned to the
        scale set.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.list_virtual_machine_scale_set_vm_network_interfaces testset testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        nics = __utils__["azurearm.paged_object_to_list"](
            netconn.network_interfaces.list_virtual_machine_scale_set_network_interfaces(
                virtual_machine_scale_set_name=scale_set,
                resource_group_name=resource_group,
            )
        )

        for nic in nics:
            result[nic["name"]] = nic
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


# pylint: disable=invalid-name
@_deprecation_message
def get_virtual_machine_scale_set_network_interface(
    name, scale_set, vm_index, resource_group, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Get information about a specific network interface within a scale set.

    :param name: The name of the network interface to query.

    :param scale_set: The name of the scale set containing the interface.

    :param vm_index: The virtual machine index.

    :param resource_group: The resource group name assigned to the
        scale set.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.get_virtual_machine_scale_set_network_interface test-iface0 testset testvm testgroup

    """
    expand = kwargs.get("expand")

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        nic = netconn.network_interfaces.list_virtual_machine_scale_set_vm_network_interfaces(
            network_interface_name=name,
            virtual_machine_scale_set_name=scale_set,
            virtualmachine_index=vm_index,
            resource_group_name=resource_group,
            exapnd=expand,
        )

        result = nic.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def public_ip_address_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a public IP address.

    :param name: The name of the public IP address to delete.

    :param resource_group: The resource group name assigned to the
        public IP address.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.public_ip_address_delete test-pub-ip testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        pub_ip = netconn.public_ip_addresses.delete(
            public_ip_address_name=name, resource_group_name=resource_group
        )
        pub_ip.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def public_ip_address_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific public IP address.

    :param name: The name of the public IP address to query.

    :param resource_group: The resource group name assigned to the
        public IP address.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.public_ip_address_get test-pub-ip testgroup

    """
    expand = kwargs.get("expand")

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        pub_ip = netconn.public_ip_addresses.get(
            public_ip_address_name=name,
            resource_group_name=resource_group,
            expand=expand,
        )
        result = pub_ip.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def public_ip_address_create_or_update(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Create or update a public IP address within a specified resource group.

    :param name: The name of the public IP address to create.

    :param resource_group: The resource group name assigned to the
        public IP address.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.public_ip_address_create_or_update test-ip-0 testgroup

    """
    if "location" not in kwargs:
        rg_props = __salt__["azurearm_resource.resource_group_get"](
            resource_group, **kwargs
        )

        if "error" in rg_props:
            log.error("Unable to determine location from resource group specified.")
            return False
        kwargs["location"] = rg_props["location"]

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        pub_ip_model = __utils__["azurearm.create_object_model"](
            "network", "PublicIPAddress", **kwargs
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        ip = netconn.public_ip_addresses.create_or_update(
            resource_group_name=resource_group,
            public_ip_address_name=name,
            parameters=pub_ip_model,
        )
        ip.wait()
        ip_result = ip.result()
        result = ip_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def public_ip_addresses_list_all(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all public IP addresses within a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.public_ip_addresses_list_all

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        pub_ips = __utils__["azurearm.paged_object_to_list"](
            netconn.public_ip_addresses.list_all()
        )

        for ip in pub_ips:
            result[ip["name"]] = ip
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def public_ip_addresses_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all public IP addresses within a resource group.

    :param resource_group: The resource group name to list public IP
        addresses within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.public_ip_addresses_list testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        pub_ips = __utils__["azurearm.paged_object_to_list"](
            netconn.public_ip_addresses.list(resource_group_name=resource_group)
        )

        for ip in pub_ips:
            result[ip["name"]] = ip
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_filter_rule_delete(name, route_filter, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a route filter rule.

    :param name: The route filter rule to delete.

    :param route_filter: The route filter containing the rule.

    :param resource_group: The resource group name assigned to the
        route filter.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_filter_rule_delete test-rule test-filter testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        rule = netconn.route_filter_rules.delete(
            resource_group_name=resource_group,
            route_filter_name=route_filter,
            rule_name=name,
        )
        rule.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def route_filter_rule_get(name, route_filter, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific route filter rule.

    :param name: The route filter rule to query.

    :param route_filter: The route filter containing the rule.

    :param resource_group: The resource group name assigned to the
        route filter.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_filter_rule_get test-rule test-filter testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        rule = netconn.route_filter_rules.get(
            resource_group_name=resource_group,
            route_filter_name=route_filter,
            rule_name=name,
        )

        result = rule.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_filter_rule_create_or_update(
    name, access, communities, route_filter, resource_group, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Create or update a rule within a specified route filter.

    :param name: The name of the rule to create.

    :param access: The access type of the rule. Valid values are 'Allow' and 'Deny'.

    :param communities: A list of BGP communities to filter on.

    :param route_filter: The name of the route filter containing the rule.

    :param resource_group: The resource group name assigned to the
        route filter.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_filter_rule_create_or_update \
                  test-rule allow "['12076:51006']" test-filter testgroup

    """
    if not isinstance(communities, list):
        log.error("The communities parameter must be a list of strings!")
        return False

    if "location" not in kwargs:
        rg_props = __salt__["azurearm_resource.resource_group_get"](
            resource_group, **kwargs
        )

        if "error" in rg_props:
            log.error("Unable to determine location from resource group specified.")
            return False
        kwargs["location"] = rg_props["location"]

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        rule_model = __utils__["azurearm.create_object_model"](
            "network",
            "RouteFilterRule",
            access=access,
            communities=communities,
            **kwargs,
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        rule = netconn.route_filter_rules.create_or_update(
            resource_group_name=resource_group,
            route_filter_name=route_filter,
            rule_name=name,
            route_filter_rule_parameters=rule_model,
        )
        rule.wait()
        rule_result = rule.result()
        result = rule_result.as_dict()
    except CloudError as exc:
        message = str(exc)
        if kwargs.get("subscription_id") == str(message).strip():
            message = "Subscription not authorized for this operation!"
        __utils__["azurearm.log_cloud_error"]("network", message, **kwargs)
        result = {"error": message}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def route_filter_rules_list(route_filter, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all routes within a route filter.

    :param route_filter: The route filter to query.

    :param resource_group: The resource group name assigned to the
        route filter.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_filter_rules_list test-filter testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        rules = __utils__["azurearm.paged_object_to_list"](
            netconn.route_filter_rules.list_by_route_filter(
                resource_group_name=resource_group, route_filter_name=route_filter
            )
        )

        for rule in rules:
            result[rule["name"]] = rule
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_filter_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a route filter.

    :param name: The name of the route filter to delete.

    :param resource_group: The resource group name assigned to the
        route filter.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_filter_delete test-filter testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        route_filter = netconn.route_filters.delete(
            route_filter_name=name, resource_group_name=resource_group
        )
        route_filter.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def route_filter_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific route filter.

    :param name: The name of the route table to query.

    :param resource_group: The resource group name assigned to the
        route filter.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_filter_get test-filter testgroup

    """
    expand = kwargs.get("expand")

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        route_filter = netconn.route_filters.get(
            route_filter_name=name, resource_group_name=resource_group, expand=expand
        )
        result = route_filter.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_filter_create_or_update(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Create or update a route filter within a specified resource group.

    :param name: The name of the route filter to create.

    :param resource_group: The resource group name assigned to the
        route filter.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_filter_create_or_update test-filter testgroup

    """
    if "location" not in kwargs:
        rg_props = __salt__["azurearm_resource.resource_group_get"](
            resource_group, **kwargs
        )

        if "error" in rg_props:
            log.error("Unable to determine location from resource group specified.")
            return False
        kwargs["location"] = rg_props["location"]

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        rt_filter_model = __utils__["azurearm.create_object_model"](
            "network", "RouteFilter", **kwargs
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        rt_filter = netconn.route_filters.create_or_update(
            resource_group_name=resource_group,
            route_filter_name=name,
            route_filter_parameters=rt_filter_model,
        )
        rt_filter.wait()
        rt_result = rt_filter.result()
        result = rt_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def route_filters_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all route filters within a resource group.

    :param resource_group: The resource group name to list route
        filters within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_filters_list testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        filters = __utils__["azurearm.paged_object_to_list"](
            netconn.route_filters.list_by_resource_group(
                resource_group_name=resource_group
            )
        )

        for route_filter in filters:
            result[route_filter["name"]] = route_filter
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_filters_list_all(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all route filters within a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_filters_list_all

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        filters = __utils__["azurearm.paged_object_to_list"](
            netconn.route_filters.list()
        )

        for route_filter in filters:
            result[route_filter["name"]] = route_filter
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_delete(name, route_table, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a route from a route table.

    :param name: The route to delete.

    :param route_table: The route table containing the route.

    :param resource_group: The resource group name assigned to the
        route table.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_delete test-rt test-rt-table testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        route = netconn.routes.delete(
            resource_group_name=resource_group,
            route_table_name=route_table,
            route_name=name,
        )
        route.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def route_get(name, route_table, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific route.

    :param name: The route to query.

    :param route_table: The route table containing the route.

    :param resource_group: The resource group name assigned to the
        route table.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_get test-rt test-rt-table testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        route = netconn.routes.get(
            resource_group_name=resource_group,
            route_table_name=route_table,
            route_name=name,
        )

        result = route.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_create_or_update(
    name,
    address_prefix,
    next_hop_type,
    route_table,
    resource_group,
    next_hop_ip_address=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Create or update a route within a specified route table.

    :param name: The name of the route to create.

    :param address_prefix: The destination CIDR to which the route applies.

    :param next_hop_type: The type of Azure hop the packet should be sent to. Possible values are:
        'VirtualNetworkGateway', 'VnetLocal', 'Internet', 'VirtualAppliance', and 'None'.

    :param next_hop_ip_address: Optional IP address to which packets should be forwarded. Next hop
        values are only allowed in routes where the next_hop_type is 'VirtualAppliance'.

    :param route_table: The name of the route table containing the route.

    :param resource_group: The resource group name assigned to the
        route table.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_create_or_update test-rt '10.0.0.0/8' test-rt-table testgroup

    """
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        rt_model = __utils__["azurearm.create_object_model"](
            "network",
            "Route",
            address_prefix=address_prefix,
            next_hop_type=next_hop_type,
            next_hop_ip_address=next_hop_ip_address,
            **kwargs,
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        route = netconn.routes.create_or_update(
            resource_group_name=resource_group,
            route_table_name=route_table,
            route_name=name,
            route_parameters=rt_model,
        )
        route.wait()
        rt_result = route.result()
        result = rt_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def routes_list(route_table, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all routes within a route table.

    :param route_table: The route table to query.

    :param resource_group: The resource group name assigned to the
        route table.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.routes_list test-rt-table testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        routes = __utils__["azurearm.paged_object_to_list"](
            netconn.routes.list(
                resource_group_name=resource_group, route_table_name=route_table
            )
        )

        for route in routes:
            result[route["name"]] = route
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_table_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete a route table.

    :param name: The name of the route table to delete.

    :param resource_group: The resource group name assigned to the
        route table.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_table_delete test-rt-table testgroup

    """
    result = False
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        table = netconn.route_tables.delete(
            route_table_name=name, resource_group_name=resource_group
        )
        table.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)

    return result


@_deprecation_message
def route_table_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get details about a specific route table.

    :param name: The name of the route table to query.

    :param resource_group: The resource group name assigned to the
        route table.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_table_get test-rt-table testgroup

    """
    expand = kwargs.get("expand")

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        table = netconn.route_tables.get(
            route_table_name=name, resource_group_name=resource_group, expand=expand
        )
        result = table.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_table_create_or_update(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Create or update a route table within a specified resource group.

    :param name: The name of the route table to create.

    :param resource_group: The resource group name assigned to the
        route table.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_table_create_or_update test-rt-table testgroup

    """
    if "location" not in kwargs:
        rg_props = __salt__["azurearm_resource.resource_group_get"](
            resource_group, **kwargs
        )

        if "error" in rg_props:
            log.error("Unable to determine location from resource group specified.")
            return False
        kwargs["location"] = rg_props["location"]

    netconn = __utils__["azurearm.get_client"]("network", **kwargs)

    try:
        rt_tbl_model = __utils__["azurearm.create_object_model"](
            "network", "RouteTable", **kwargs
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        table = netconn.route_tables.create_or_update(
            resource_group_name=resource_group,
            route_table_name=name,
            parameters=rt_tbl_model,
        )
        table.wait()
        tbl_result = table.result()
        result = tbl_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def route_tables_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all route tables within a resource group.

    :param resource_group: The resource group name to list route
        tables within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_tables_list testgroup

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        tables = __utils__["azurearm.paged_object_to_list"](
            netconn.route_tables.list(resource_group_name=resource_group)
        )

        for table in tables:
            result[table["name"]] = table
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def route_tables_list_all(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all route tables within a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_network.route_tables_list_all

    """
    result = {}
    netconn = __utils__["azurearm.get_client"]("network", **kwargs)
    try:
        tables = __utils__["azurearm.paged_object_to_list"](
            netconn.route_tables.list_all()
        )

        for table in tables:
            result[table["name"]] = table
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("network", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result



import hashlib
import os
import stat
import time
from pathlib import Path

import pytest
import zmq.eventloop.ioloop

import salt.config
import salt.ext.tornado.ioloop
import salt.ext.tornado.iostream
import salt.utils.event
import salt.utils.stringutils
from salt.utils.event import SaltEvent
from tests.support.events import eventpublisher_process, eventsender_process
from tests.support.mock import patch

NO_LONG_IPC = False
if getattr(zmq, "IPC_PATH_MAX_LEN", 103) <= 103:
    NO_LONG_IPC = True

pytestmark = [
    pytest.mark.skipif(
        NO_LONG_IPC,
        reason="This system does not support long IPC paths. Skipping event tests!",
    ),
]


@pytest.fixture(autouse=True)
def sock_dir(tmp_path):
    sock_dir_path = tmp_path / "test-socks"
    sock_dir_path.mkdir(parents=True, exist_ok=True)
    yield sock_dir_path


def _assert_got_event(evt, data, msg=None, expected_failure=False):
    assert evt is not None, msg
    for key in data:
        assert key in evt, f"{msg}: Key {key} missing"
        assertMsg = "{0}: Key {1} value mismatch, {2} != {3}"
        assertMsg = assertMsg.format(msg, key, data[key], evt[key])
        if not expected_failure:
            assert data[key] == evt[key], assertMsg
        else:
            assert data[key] != evt[key]


def test_master_event(sock_dir):
    with salt.utils.event.MasterEvent(str(sock_dir), listen=False) as me:
        assert me.puburi == str(sock_dir / "master_event_pub.ipc")
        assert me.pulluri == str(sock_dir / "master_event_pull.ipc")


def test_minion_event(sock_dir):
    opts = dict(id="foo", sock_dir=str(sock_dir))
    id_hash = hashlib.sha256(salt.utils.stringutils.to_bytes(opts["id"])).hexdigest()[
        :10
    ]
    with salt.utils.event.MinionEvent(opts, listen=False) as me:
        assert me.puburi == str(sock_dir / f"minion_event_{id_hash}_pub.ipc")
        assert me.pulluri == str(sock_dir / f"minion_event_{id_hash}_pull.ipc")


def test_minion_event_tcp_ipc_mode():
    opts = dict(id="foo", ipc_mode="tcp")
    with salt.utils.event.MinionEvent(opts, listen=False) as me:
        assert me.puburi == 4510
        assert me.pulluri == 4511


def test_minion_event_no_id(sock_dir):
    with salt.utils.event.MinionEvent(dict(sock_dir=str(sock_dir)), listen=False) as me:
        id_hash = hashlib.sha256(salt.utils.stringutils.to_bytes("")).hexdigest()[:10]
        assert me.puburi == str(sock_dir / f"minion_event_{id_hash}_pub.ipc")
        assert me.pulluri == str(sock_dir / f"minion_event_{id_hash}_pull.ipc")


@pytest.mark.slow_test
def test_event_single(sock_dir):
    """Test a single event is received"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            evt1 = me.get_event(tag="evt1")
            _assert_got_event(evt1, {"data": "foo1"})


@pytest.mark.slow_test
def test_event_single_no_block(sock_dir):
    """Test a single event is received, no block"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            start = time.time()
            finish = start + 5
            evt1 = me.get_event(wait=0, tag="evt1", no_block=True)
            # We should get None and way before the 5 seconds wait since it's
            # non-blocking, otherwise it would wait for an event which we
            # didn't even send
            assert evt1 is None, None
            assert start < finish
            me.fire_event({"data": "foo1"}, "evt1")
            evt1 = me.get_event(wait=0, tag="evt1")
            _assert_got_event(evt1, {"data": "foo1"})


@pytest.mark.slow_test
def test_event_single_wait_0_no_block_False(sock_dir):
    """Test a single event is received with wait=0 and no_block=False and doesn't spin the while loop"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            # This is too fast and will be None but assures we're not blocking
            evt1 = me.get_event(wait=0, tag="evt1", no_block=False)
            _assert_got_event(evt1, {"data": "foo1"})


@pytest.mark.slow_test
def test_event_timeout(sock_dir):
    """Test no event is received if the timeout is reached"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            evt1 = me.get_event(tag="evt1")
            _assert_got_event(evt1, {"data": "foo1"})
            evt2 = me.get_event(tag="evt1")
            assert evt2 is None


@pytest.mark.slow_test
def test_event_no_timeout(sock_dir):
    """Test no wait timeout, we should block forever, until we get one"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            with eventsender_process({"data": "foo2"}, "evt2", str(sock_dir), 5):
                evt = me.get_event(tag="evt2", wait=0, no_block=False)
            _assert_got_event(evt, {"data": "foo2"})


@pytest.mark.slow_test
def test_event_matching(sock_dir):
    """Test a startswith match"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            evt1 = me.get_event(tag="ev")
            _assert_got_event(evt1, {"data": "foo1"})


@pytest.mark.slow_test
def test_event_matching_regex(sock_dir):
    """Test a regex match"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            evt1 = me.get_event(tag="^ev", match_type="regex")
            _assert_got_event(evt1, {"data": "foo1"})


@pytest.mark.slow_test
def test_event_matching_all(sock_dir):
    """Test an all match"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            evt1 = me.get_event(tag="")
            _assert_got_event(evt1, {"data": "foo1"})


@pytest.mark.slow_test
def test_event_matching_all_when_tag_is_None(sock_dir):
    """Test event matching all when not passing a tag"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            evt1 = me.get_event()
            _assert_got_event(evt1, {"data": "foo1"})


@pytest.mark.slow_test
def test_event_not_subscribed(sock_dir):
    """Test get_event drops non-subscribed events"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            me.fire_event({"data": "foo2"}, "evt2")
            evt2 = me.get_event(tag="evt2")
            evt1 = me.get_event(tag="evt1")
            _assert_got_event(evt2, {"data": "foo2"})
            assert evt1 is None


@pytest.mark.slow_test
def test_event_subscription_cache(sock_dir):
    """Test subscriptions cache a message until requested"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.subscribe("evt1")
            me.fire_event({"data": "foo1"}, "evt1")
            me.fire_event({"data": "foo2"}, "evt2")
            evt2 = me.get_event(tag="evt2")
            evt1 = me.get_event(tag="evt1")
            _assert_got_event(evt2, {"data": "foo2"})
            _assert_got_event(evt1, {"data": "foo1"})


@pytest.mark.slow_test
def test_event_subscriptions_cache_regex(sock_dir):
    """Test regex subscriptions cache a message until requested"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.subscribe("e..1$", "regex")
            me.fire_event({"data": "foo1"}, "evt1")
            me.fire_event({"data": "foo2"}, "evt2")
            evt2 = me.get_event(tag="evt2")
            evt1 = me.get_event(tag="evt1")
            _assert_got_event(evt2, {"data": "foo2"})
            _assert_got_event(evt1, {"data": "foo1"})


@pytest.mark.slow_test
def test_event_multiple_clients(sock_dir):
    """Test event is received by multiple clients"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(
            str(sock_dir), listen=True
        ) as me1, salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me2:
            # We need to sleep here to avoid a race condition wherein
            # the second socket may not be connected by the time the first socket
            # sends the event.
            time.sleep(0.5)
            me1.fire_event({"data": "foo1"}, "evt1")
            evt1 = me1.get_event(tag="evt1")
            _assert_got_event(evt1, {"data": "foo1"})
            evt2 = me2.get_event(tag="evt1")
            _assert_got_event(evt2, {"data": "foo1"})


def test_event_nested_sub_all(sock_dir):
    """Test nested event subscriptions do not drop events, get event for all tags"""
    # Show why not to call get_event(tag='')
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            me.fire_event({"data": "foo2"}, "evt2")
            evt2 = me.get_event(tag="")
            evt1 = me.get_event(tag="")
            _assert_got_event(evt2, {"data": "foo2"}, expected_failure=True)
            _assert_got_event(evt1, {"data": "foo1"}, expected_failure=True)


@pytest.mark.slow_test
def test_event_many(sock_dir):
    """Test a large number of events, one at a time"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            for i in range(500):
                me.fire_event({"data": f"{i}"}, "testevents")
                evt = me.get_event(tag="testevents")
                _assert_got_event(evt, {"data": f"{i}"}, f"Event {i}")


@pytest.mark.slow_test
def test_event_many_backlog(sock_dir):
    """Test a large number of events, send all then recv all"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            # Must not exceed zmq HWM
            for i in range(500):
                me.fire_event({"data": f"{i}"}, "testevents")
            for i in range(500):
                evt = me.get_event(tag="testevents")
                _assert_got_event(evt, {"data": f"{i}"}, f"Event {i}")


# Test the fire_master function. As it wraps the underlying fire_event,
# we don't need to perform extensive testing.
@pytest.mark.slow_test
def test_send_master_event(sock_dir):
    """Tests that sending an event through fire_master generates expected event"""
    with eventpublisher_process(str(sock_dir)):
        with salt.utils.event.MasterEvent(str(sock_dir), listen=True) as me:
            data = {"data": "foo1"}
            me.fire_master(data, "test_master")

            evt = me.get_event(tag="fire_master")
            _assert_got_event(
                evt,
                {
                    "data": data,
                    "tag": "test_master",
                    "events": None,
                    "pretag": None,
                },
            )


def test_connect_pull_should_debug_log_on_StreamClosedError():
    event = SaltEvent(node=None)
    with patch.object(event, "pusher") as mock_pusher:
        with patch.object(
            salt.utils.event.log, "debug", autospec=True
        ) as mock_log_debug:
            mock_pusher.connect.side_effect = (
                salt.ext.tornado.iostream.StreamClosedError
            )
            event.connect_pull()
            call = mock_log_debug.mock_calls[0]
            assert call.args[0] == "Unable to connect pusher: %s"
            assert isinstance(call.args[1], salt.ext.tornado.iostream.StreamClosedError)
            assert call.args[1].args[0] == "Stream is closed"


@pytest.mark.parametrize("error", [Exception, KeyError, IOError])
def test_connect_pull_should_error_log_on_other_errors(error):
    event = SaltEvent(node=None)
    with patch.object(event, "pusher") as mock_pusher:
        with patch.object(
            salt.utils.event.log, "debug", autospec=True
        ) as mock_log_debug:
            with patch.object(
                salt.utils.event.log, "error", autospec=True
            ) as mock_log_error:
                mock_pusher.connect.side_effect = error
                event.connect_pull()
                mock_log_debug.assert_not_called()
                call = mock_log_error.mock_calls[0]
                assert call.args[0] == "Unable to connect pusher: %s"
                assert not isinstance(
                    call.args[1], salt.ext.tornado.iostream.StreamClosedError
                )


@pytest.mark.slow_test
def test_master_pub_permissions(sock_dir):
    with eventpublisher_process(str(sock_dir)):
        p = Path(str(sock_dir)) / "master_event_pub.ipc"
        mode = os.lstat(p).st_mode
        assert bool(os.lstat(p).st_mode & stat.S_IRUSR)
        assert not bool(os.lstat(p).st_mode & stat.S_IRGRP)
        assert not bool(os.lstat(p).st_mode & stat.S_IROTH)



"""
NAPALM: Network Automation and Programmability Abstraction Layer with Multivendor support
=========================================================================================

.. versionadded:: 2016.11.0

Proxy minion for managing network devices via NAPALM_ library.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net> & Jerome Fleury <jf@cloudflare.com>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------

The ``napalm`` proxy module requires NAPALM_ library to be installed:  ``pip install napalm``
Please check Installation_ for complete details.

.. _NAPALM: https://napalm-automation.net/
.. _Installation: http://napalm.readthedocs.io/en/latest/installation/index.html

.. note::

    Beginning with Salt release 2017.7.3, it is recommended to use
    ``napalm`` >= ``2.0.0``. The library has been unified into a monolithic
    package, as in opposite to separate packages per driver. For more details
    you can check `this document <https://napalm-automation.net/reunification/>`_.
    While it will still work with the old packages, bear in mind that the NAPALM
    core team will maintain only the main ``napalm`` package.

    Moreover, for additional capabilities, the users can always define a
    library that extends NAPALM's base capabilities and configure the
    ``provider`` option (see below).

Pillar
------

The napalm proxy configuration requires the following parameters in order to connect to the network device:

driver
    Specifies the network device operating system.
    For a complete list of the supported operating systems please refer to the
    `NAPALM Read the Docs page`_.

host
    The IP Address or FQDN to use when connecting to the device. Alternatively,
    the following field names can be used instead: ``hostname``, ``fqdn``, ``ip``.

username
    The username to be used when connecting to the device.

passwd
    The password needed to establish the connection.

    .. note::

        This field may not be mandatory when working with SSH-based drivers, and
        the username has a SSH key properly configured on the device targeted to
        be managed.

optional_args
    Dictionary with the optional arguments.
    Check the complete list of supported `optional arguments`_.

always_alive: ``True``
    In certain less dynamic environments, maintaining the remote connection permanently
    open with the network device is not always beneficial. In that case, the user can
    select to initialize the connection only when needed, by specifying this field to ``false``.
    Default: ``true`` (maintains the connection with the remote network device).

    .. versionadded:: 2017.7.0

provider: ``napalm_base``
    The library that provides the ``get_network_device`` function.
    This option is useful when the user has more specific needs and requires
    to extend the NAPALM capabilities using a private library implementation.
    The only constraint is that the alternative library needs to have the
    ``get_network_device`` function available.

    .. versionadded:: 2017.7.1

multiprocessing: ``False``
    Overrides the :conf_minion:`multiprocessing` option, per proxy minion.
    The ``multiprocessing`` option must be turned off for SSH-based proxies.
    However, some NAPALM drivers (e.g. Arista, NX-OS) are not SSH-based.
    As multiple proxy minions may share the same configuration file,
    this option permits the configuration of the ``multiprocessing`` option
    more specifically, for some proxy minions.

    .. versionadded:: 2017.7.2


.. _`NAPALM Read the Docs page`: https://napalm.readthedocs.io/en/latest/#supported-network-operating-systems
.. _`optional arguments`: http://napalm.readthedocs.io/en/latest/support/index.html#list-of-supported-optional-arguments

Proxy pillar file example:

.. code-block:: yaml

    proxy:
      proxytype: napalm
      driver: junos
      host: core05.nrt02
      username: my_username
      passwd: my_password
      optional_args:
        port: 12201

Example using a user-specific library, extending NAPALM's capabilities, e.g. ``custom_napalm_base``:

.. code-block:: yaml

    proxy:
      proxytype: napalm
      driver: ios
      fqdn: cr1.th2.par.as1234.net
      username: salt
      password: ''
      provider: custom_napalm_base

.. seealso::

    - :mod:`NAPALM grains: select network devices based on their characteristics <salt.grains.napalm>`
    - :mod:`NET module: network basic features <salt.modules.napalm_network>`
    - :mod:`Network config state: Manage the configuration using arbitrary templates <salt.states.netconfig>`
    - :mod:`NAPALM YANG state: Manage the configuration according to the YANG models (OpenConfig/IETF) <salt.states.netyang>`
    - :mod:`Network ACL module: Generate and load ACL (firewall) configuration <salt.modules.napalm_acl>`
    - :mod:`Network ACL state: Manage the firewall configuration <salt.states.netacl>`
    - :mod:`NTP operational and configuration management module <salt.modules.napalm_ntp>`
    - :mod:`BGP operational and configuration management module <salt.modules.napalm_bgp>`
    - :mod:`Routes details <salt.modules.napalm_route>`
    - :mod:`SNMP configuration module <salt.modules.napalm_snmp>`
    - :mod:`Users configuration management <salt.modules.napalm_users>`

.. note::
    Beginning with release codename 2019.2.0, any NAPALM command executed when
    running under a NAPALM Proxy Minion supports the ``force_reconnect``
    magic argument.

    Proxy Minions generally establish a connection with the remote network
    device at the time of the Minion startup and that connection is going to be
    used forever.

    If one would need execute a command on the device but connecting using
    different parameters (due to various causes, e.g., unable to authenticate
    the user specified in the Pillar as the authentication system - say
    TACACS+ is not available, or the DNS resolver is currently down and would
    like to temporarily use the IP address instead, etc.), it implies updating
    the Pillar data and restarting the Proxy Minion process restart.
    In particular cases like that, you can pass the ``force_reconnect=True``
    keyword argument, together with the alternative connection details, to
    enforce the command to be executed over a separate connection.

    For example, if the usual command is ``salt '*' net.arp``, you can use the
    following to connect using a different username instead:
    ``salt '*' net.arp username=my-alt-usr force_reconnect=True``.
"""

import logging

import salt.utils.napalm

log = logging.getLogger(__file__)


# ----------------------------------------------------------------------------------------------------------------------
# proxy properties
# ----------------------------------------------------------------------------------------------------------------------

__proxyenabled__ = ["napalm"]
# proxy name


# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    return salt.utils.napalm.virtual(__opts__, "napalm", __file__)


# ----------------------------------------------------------------------------------------------------------------------
# helper functions -- will not be exported
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# Proxy functions
# ----------------------------------------------------------------------------------------------------------------------


def init(opts):
    """
    Opens the connection with the network device.
    """
    napalm_device = salt.utils.napalm.get_device(opts)
    __context__["napalm_device"] = {
        "opts": opts,
        "id": opts["id"],
        "network_device": napalm_device,
        "details": {"initialized": True},
    }
    return True


def alive(opts):
    """
    Return the connection status with the remote device.

    .. versionadded:: 2017.7.0
    """
    if salt.utils.napalm.not_always_alive(opts):
        return True  # don't force reconnection for not-always alive proxies
        # or regular minion
    is_alive_ret = call("is_alive", **{})
    if not is_alive_ret.get("result", False):
        log.debug(
            "[%s] Unable to execute `is_alive`: %s",
            opts.get("id"),
            is_alive_ret.get("comment"),
        )
        # if `is_alive` is not implemented by the underneath driver,
        # will consider the connection to be still alive
        # we don't want overly request connection reestablishment
        # NOTE: revisit this if IOS is still not stable
        #       and return False to force reconnection
        return True
    flag = is_alive_ret.get("out", {}).get("is_alive", False)
    log.debug("Is %s still alive? %s", opts.get("id"), "Yes." if flag else "No.")
    return flag


def ping():
    """
    Connection open successfully?
    """
    return (
        __context__.get("napalm_device", {}).get("network_device", {}).get("UP", False)
    )


def initialized():
    """
    Connection finished initializing?
    """
    return (
        __context__.get("napalm_device", {})
        .get("details", {})
        .get("initialized", False)
    )


def get_device():
    """
    Returns the network device object.
    """
    return __context__.get("napalm_device", {}).get("network_device", {})


def get_grains():
    """
    Retrieve facts from the network device.
    """
    return call("get_facts", **{})


def grains_refresh():
    """
    Refresh the grains.
    """
    __context__["napalm_device"]["details"]["grains_cache"] = {}
    return get_grains()


def fns():
    """
    Method called by NAPALM grains module.
    """
    return {"details": "Network device grains."}


def shutdown(opts):
    """
    Closes connection with the device.
    """
    try:
        if not __context__["napalm_device"]["network_device"].get("UP", False):
            raise Exception("not connected!")
        __context__["napalm_device"]["network_device"].get("DRIVER").close()
    except Exception as error:  # pylint: disable=broad-except
        port = (
            __context__["napalm_device"]["network_device"]
            .get("OPTIONAL_ARGS", {})
            .get("port"),
        )
        log.error(
            "Cannot close connection with %s%s! Please check error: %s",
            __context__["napalm_device"]["network_device"].get(
                "HOSTNAME", "[unknown hostname]"
            ),
            f":{port}" if port else "",
            error,
        )

    return True


# ----------------------------------------------------------------------------------------------------------------------
# Callable functions
# ----------------------------------------------------------------------------------------------------------------------


def call(method, *args, **kwargs):
    """
    Calls a specific method from the network driver instance.
    Please check the readthedocs_ page for the updated list of getters.

    .. _readthedocs: http://napalm.readthedocs.org/en/latest/support/index.html#getters-support-matrix

    :param method: specifies the name of the method to be called
    :param params: contains the mapping between the name and the values of the parameters needed to call the method
    :return: A dictionary with three keys:

    - result (True/False): if the operation succeeded
    - out (object): returns the object as-is from the call
    - comment (string): provides more details in case the call failed
    - traceback (string): complete traceback in case of exception. Please
      submit an issue including this traceback on the `correct driver repo`_
      and make sure to read the FAQ_

    .. _`correct driver repo`: https://github.com/napalm-automation/napalm/issues/new
    .. _FAQ: https://github.com/napalm-automation/napalm#faq

    Example:

    .. code-block:: python

        __proxy__['napalm.call']('cli'
                                 **{
                                    'commands': [
                                        'show version',
                                        'show chassis fan'
                                    ]
                                 })
    """
    kwargs_copy = {}
    kwargs_copy.update(kwargs)
    for karg, warg in kwargs_copy.items():
        # will remove None values
        # thus the NAPALM methods will be called with their defaults
        if warg is None:
            kwargs.pop(karg)
    if "napalm_device" in __context__:
        return salt.utils.napalm.call(
            __context__["napalm_device"]["network_device"], method, *args, **kwargs
        )
    else:
        return {"result": False, "comment": "Not initialised yet", "out": None}



"""
Integration tests for timezone module

Linux and Solaris are supported
"""

import subprocess

import pytest

import salt.utils.platform
from tests.support.case import ModuleCase

try:
    import tzlocal  # pylint: disable=unused-import

    HAS_TZLOCAL = True
except ImportError:
    HAS_TZLOCAL = False


def _check_systemctl():
    if not hasattr(_check_systemctl, "memo"):
        if not salt.utils.platform.is_linux():
            _check_systemctl.memo = False
        else:
            proc = subprocess.run(["timedatectl"], capture_output=True, check=False)
            _check_systemctl.memo = b"No such file or directory" in proc.stderr
    return _check_systemctl.memo


@pytest.mark.skipif(_check_systemctl(), reason="systemctl degraded")
class TimezoneLinuxModuleTest(ModuleCase):
    def setUp(self):
        """
        Set up Linux test environment
        """
        ret_grain = self.run_function("grains.item", ["kernel"])
        if "Linux" not in ret_grain["kernel"]:
            self.skipTest("For Linux only")
        super().setUp()

    def test_get_hwclock(self):
        timescale = ["UTC", "localtime"]
        ret = self.run_function("timezone.get_hwclock")
        self.assertIn(ret, timescale)


@pytest.mark.skipif(_check_systemctl(), reason="systemctl degraded")
class TimezoneSolarisModuleTest(ModuleCase):
    def setUp(self):
        """
        Set up Solaris test environment
        """
        ret_grain = self.run_function("grains.item", ["os_family"])
        if "Solaris" not in ret_grain["os_family"]:
            self.skipTest("For Solaris only")
        super().setUp()

    def test_get_hwclock(self):
        timescale = ["UTC", "localtime"]
        ret = self.run_function("timezone.get_hwclock")
        self.assertIn(ret, timescale)


@pytest.mark.skip_unless_on_windows
class TimezoneWindowsModuleTest(ModuleCase):
    def setUp(self):
        self.pre = self.run_function("timezone.get_zone")

    def tearDown(self):
        post = self.run_function("timezone.get_zone")
        if self.pre != post:
            self.run_function("timezone.set_zone", [self.pre])

    def test_get_hwclock(self):
        timescale = ["UTC", "localtime"]
        ret = self.run_function("timezone.get_hwclock")
        self.assertIn(ret, timescale)

    @pytest.mark.destructive_test
    def test_get_zone(self):
        """
        test timezone.set_zone, get_zone and zone_compare
        """

        zone = "America/Inuvik" if not HAS_TZLOCAL else "America/Denver"

        # first set the zone
        assert self.run_function("timezone.set_zone", [zone])

        # check it set the correct zone
        ret = self.run_function("timezone.get_zone")
        assert zone in ret

        # compare zones
        assert self.run_function("timezone.zone_compare", [zone])

    def test_get_offset(self):
        """
        test timezone.get_offset
        """
        ret = self.run_function("timezone.get_offset")
        self.assertIn("-", ret)



# Python libs
import pytest

# Salt libs
import salt.beacons.smartos_imgadm as imgadm
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {imgadm: {"__context__": {}, "__salt__": {}}}


@pytest.fixture
def mock_clean_state():
    return {"first_run": True, "vms": []}


@pytest.fixture
def mock_image_none():
    return {}


@pytest.fixture
def mock_image_one():
    return {
        "00000000-0000-0000-0000-000000000001": {
            "description": "Example Image 1",
            "name": "example-1",
            "os": "smartos",
            "published": "2018-01-01T00:42:00Z",
            "source": "https://images.joyent.com",
            "version": "18.1.0",
        },
    }


@pytest.fixture
def mock_image_two():
    return {
        "00000000-0000-0000-0000-000000000001": {
            "description": "Example Image 1",
            "name": "example-1",
            "os": "smartos",
            "published": "2018-01-01T00:42:00Z",
            "source": "https://images.joyent.com",
            "version": "18.1.0",
        },
        "00000000-0000-0000-0000-000000000002": {
            "description": "Example Image 2",
            "name": "example-2",
            "os": "smartos",
            "published": "2018-01-01T00:42:00Z",
            "source": "https://images.joyent.com",
            "version": "18.2.0",
        },
    }


def test_non_list_config():
    """
    We only have minimal validation so we test that here
    """
    assert imgadm.validate({}) == (
        False,
        "Configuration for imgadm beacon must be a list!",
    )


def test_imported_startup(mock_clean_state, mock_image_one):
    """
    Test with one image and startup_import_event
    """
    # NOTE: this should yield 1 imported event
    with patch.dict(imgadm.IMGADM_STATE, mock_clean_state), patch.dict(
        imgadm.__salt__, {"imgadm.list": MagicMock(return_value=mock_image_one)}
    ):

        config = [{"startup_import_event": True}]
        assert imgadm.validate(config) == (True, "Valid beacon configuration")

        ret = imgadm.beacon(config)
        res = [
            {
                "description": "Example Image 1",
                "name": "example-1",
                "os": "smartos",
                "published": "2018-01-01T00:42:00Z",
                "source": "https://images.joyent.com",
                "tag": "imported/00000000-0000-0000-0000-000000000001",
                "version": "18.1.0",
            }
        ]
        assert ret == res


def test_imported_nostartup(mock_clean_state, mock_image_one):
    """
    Test with one image and startup_import_event unset/false
    """
    # NOTE: this should yield 0 imported event
    with patch.dict(imgadm.IMGADM_STATE, mock_clean_state), patch.dict(
        imgadm.__salt__, {"imgadm.list": MagicMock(return_value=mock_image_one)}
    ):

        config = []

        assert imgadm.validate(config) == (True, "Valid beacon configuration")
        assert imgadm.beacon(config) == []


def test_imported(mock_clean_state, mock_image_one, mock_image_two):
    """
    Test with one image and a new image added on the 2nd pass
    """
    # NOTE: this should yield 1 imported event
    with patch.dict(imgadm.IMGADM_STATE, mock_clean_state), patch.dict(
        imgadm.__salt__,
        {"imgadm.list": MagicMock(side_effect=[mock_image_one, mock_image_two])},
    ):

        config = []
        assert imgadm.validate(config) == (True, "Valid beacon configuration")

        # Initial pass (Initialized state and do not yield imported images at startup)
        imgadm.beacon(config)

        # Second pass (After importing a new image)
        ret = imgadm.beacon(config)
        res = [
            {
                "description": "Example Image 2",
                "name": "example-2",
                "os": "smartos",
                "published": "2018-01-01T00:42:00Z",
                "source": "https://images.joyent.com",
                "tag": "imported/00000000-0000-0000-0000-000000000002",
                "version": "18.2.0",
            }
        ]

        assert ret == res


def test_deleted(mock_clean_state, mock_image_one, mock_image_two):
    """
    Test with two images and one gets deletes
    """
    # NOTE: this should yield 1 deleted event
    with patch.dict(imgadm.IMGADM_STATE, mock_clean_state), patch.dict(
        imgadm.__salt__,
        {"imgadm.list": MagicMock(side_effect=[mock_image_two, mock_image_one])},
    ):

        config = []
        assert imgadm.validate(config) == (True, "Valid beacon configuration")

        # Initial pass (Initialized state and do not yield imported images at startup)
        imgadm.beacon(config)

        # Second pass (After deleting one image)
        ret = imgadm.beacon(config)
        res = [
            {
                "description": "Example Image 2",
                "name": "example-2",
                "os": "smartos",
                "published": "2018-01-01T00:42:00Z",
                "source": "https://images.joyent.com",
                "tag": "deleted/00000000-0000-0000-0000-000000000002",
                "version": "18.2.0",
            }
        ]

        assert ret == res


def test_complex(mock_clean_state, mock_image_one, mock_image_two, mock_image_none):
    """
    Test with one image, delete both, import 2
    """
    # NOTE: this should yield 1 delete and 2 import events
    with patch.dict(imgadm.IMGADM_STATE, mock_clean_state), patch.dict(
        imgadm.__salt__,
        {
            "imgadm.list": MagicMock(
                side_effect=[mock_image_one, mock_image_none, mock_image_two]
            )
        },
    ):

        config = []
        assert imgadm.validate(config), (True, "Valid beacon configuration")

        # Initial pass (Initialized state and do not yield imported images at startup)
        imgadm.beacon(config)

        # Second pass (After deleting one image)
        ret = imgadm.beacon(config)
        res = [
            {
                "description": "Example Image 1",
                "name": "example-1",
                "os": "smartos",
                "published": "2018-01-01T00:42:00Z",
                "source": "https://images.joyent.com",
                "tag": "deleted/00000000-0000-0000-0000-000000000001",
                "version": "18.1.0",
            }
        ]
        assert ret == res

        # Third pass (After importing two images)
        ret = imgadm.beacon(config)
        res = [
            {
                "description": "Example Image 1",
                "name": "example-1",
                "os": "smartos",
                "published": "2018-01-01T00:42:00Z",
                "source": "https://images.joyent.com",
                "tag": "imported/00000000-0000-0000-0000-000000000001",
                "version": "18.1.0",
            },
            {
                "description": "Example Image 2",
                "name": "example-2",
                "os": "smartos",
                "published": "2018-01-01T00:42:00Z",
                "source": "https://images.joyent.com",
                "tag": "imported/00000000-0000-0000-0000-000000000002",
                "version": "18.2.0",
            },
        ]
        assert len(ret) == 2
        for item in ret:
            assert item in res
